
import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.tools.io.*;
import org.autochthonous.graphics.rt.util.NoiseGenerator;
import org.autochthonous.graphics.rt.util.ParametersMap;
import org.autochthonous.graphics.rt.shaders.ct.*;

/**
 * A bumpy shader - an extension of the Plastic one.
 * <p>
 * Example:
 * <code>
 * <pre>
 * &lt;Bumpy diffuse="0.1" specular="0.3" roughness="0.02"
 *      transparency="0.9" refractiveIndex="1.2" scaleFactors="0.5, 0.5, 0.5"&gt;
 *      &lt;Transform&gt;
 *          &lt;Scale value="0.05, 0.05, 0.05"/&gt;
 *      &lt;/Transform&gt;
 * &lt;/Bumpy&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Bumpy (Transform)&gt;
 *     &lt;!ATTLIST Bumpy
 *         ambient              (%Color3f; | %Float;) "0, 0, 0"
 *         diffuse              (%Color3f; | %Float;) "1, 1, 1"
 *         specular             (%Color3f; | %Float;) "0, 0, 0"
 *         roughness            %Float                "0"
 *         reflectance          (%Color3f; | %Float;) "0, 0, 0"
 *         refractiveIndex      %Float                "1"
 *         transparency          (%Color3f; | %Float;) "0, 0, 0"
 *         reflectionRoughness  %Float                 "0"&gt;
 *         scaleFactors         %Vector3f;             "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class Bumpy
    extends Plastic
{
    float       xFact = 1;
    float       yFact = 1;
    float       zFact = 1;
    Vector3f    pTurb = new Vector3f();
    float[]     deltas = new float[2];
    
    NoiseGenerator    generator = NoiseGenerator.getInstance();

    public Bumpy(
        World   world)
    {
        super(world);
    }

    public static Shader getInstance(
        World         world,
        RMLReader       in)
        throws IOException
    {
        Bumpy   m = new Bumpy(world);

        m.readObject(in);

        return m;
    }

    /**
     * handle the parameters we want, passing the extras
     * to our super class for processing.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        if (params.get("scaleFactors") != null)
        {
            StringTokenizer st = new StringTokenizer((String)params.get("scaleFactors"), " ,");
            setScaleFactors(Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()));
        }
    }

    /**
     * set the scale factors for the bumpyness.
     */
    public void setScaleFactors(
        float    xFact,
        float    yFact,
        float    zFact)
    {
        this.xFact = xFact;
        this.yFact = yFact;
        this.zFact = zFact;
    }

    /**
     * the shade routine. We do the normal pertubation and then
     * pass the result back to the superclass for processing.
     */
    public void shade(
        Hit        hit,
        Pixel      p)
    {
        Point3f     loc = new Point3f(hit.oP());

        this.toShaderSpace(loc);

        p.set(0, 0, 0, 1);

        pTurb = generator.getNoiseV(loc);
        
        pTurb.x *= xFact;
        pTurb.y *= yFact;
        pTurb.z *= zFact;

        hit.perturbN(pTurb);

        super.shade(hit, p);
    }
}
