package org.autochthonous.graphics.rt;

import java.io.*;
import java.awt.*;
import java.awt.image.*;

import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * A tile.
 * <p>
 * By default tiles are repeating. Any image format supported by the JDK
 * can be used together with the ".pix" format generated by the default
 * renderer.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Tile image="tile.gif" repeating="false"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Boolean  "(true|false)"&gt;
 *     &lt;!ENTITY %String   "CDATA"&gt;
 *
 *     &lt;!ELEMENT Tile EMPTY&gt;
 *     &lt;!ATTLIST Tile
 *         image              %String;  #REQUIRED
 *         repeating          %Boolean; "true"&gt;
 * </pre>
 * </code>
 */
public class Tile
{
    protected static final int    U = 0;
    protected static final int    V = 1;

    private int[]       pix;
    private int         width;
    private int         height;

    private boolean     repeating = true;

    /**
     * default constructor, a tiling with a repeating image.
     *
     * @param im the image to be used for the tiling
     */
    public Tile(
        Image        im)
    {
        this(im, true);
    }

    /**
     * constructor, a tiling with an optionally repeating image.
     *
     * @param im the image to be used for the tiling
     * @param repeating true if we repeat, false if we don't.
     */
    public Tile(
        Image       im,
        boolean     repeating)
    {
        PixelGrabber    pg = new PixelGrabber(im, 0, 0, -1, -1, true);

        try
        {
            if (pg.grabPixels())
            {
                pix = (int[])pg.getPixels();
                width = pg.getWidth();
                height = pg.getHeight();
            }
            else
            {
                throw new RuntimeException("can't grab pixels");
            }
        }
        catch (InterruptedException e)
        {
            throw new RuntimeException(e.toString());
        }
    }

    public static Tile getInstance(
        RMLReader   in)
        throws IOException
    {
        ParametersMap   params = in.parameters();
        String          imageName = params.getString("image", null);
        boolean         repeating = params.getBoolean("repeating", "true");

        if (imageName == null)
        {
            throw new IOException("No image name specified in tile.");
        }

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return new Tile(ImageUtil.loadImage(imageName), repeating);
    }

    public Pixel getPixel(
        float       s,
        float       t)
    {
        int    sI, tI, pixel;

        if (!repeating)
        {
            if ((s > 1.0) || (s < 0.0)
                || (t > 1.0) || (t < 0.0))
            {
                return null;
            }
        }
        
        s = s - (int)s;
        t = t - (int)t;

        if (s < 0.0)
        { 
            sI = (int)(width * Math.abs(1.0 + s)) % width;
        }
        else
        {
            sI = (int)(width * s) % width;
        }

        if (t < 0.0)
        {
            tI = (int)(height * Math.abs(1.0 + t)) % height;
        }
        else
        {
            tI = (int)(height * t) % height;
        }

        pixel = pix[width * tI + sI];

        //
        // convert the int to a Pixel...
        //
        Pixel p = new Pixel();

        p.a = ((pixel >> 24) & 0xff) / 255.0f;
        p.r = ((pixel >> 16) & 0xff) / 255.0f;
        p.g = ((pixel >> 8) & 0xff) / 255.0f;
        p.b = ((pixel) & 0xff) / 255.0f;
        
        return p;
    }

    /*
     * return a grey intenstity value for a color triple.
     * (uses standard PAL values.)
     */
    private float greyscale(
        float    r,
        float    g,
        float    b)
    {
        return (r * 0.3f + g * 0.59f + b * 0.11f);
    }

    private float getR(
        int    pixel)
    {
        return ((pixel >> 16) & 0xff) / 255.0f;
    }

    private float getG(
        int    pixel)
    {
        return ((pixel >> 8) & 0xff) / 255.0f;
    }

    private float getB(
        int    pixel)
    {
        return (pixel & 0xff) / 255.0f;
    }

    /**
     * returns the greyscale differentials in s and t for the image at the
     * position s and t
     */
    public boolean getDeltas(
        float       s,
        float       t,
        float[]     deltas)
    {
        /*
         * we get indexes for a tile two less in width and height
         * and then add one so that we can always calculate a change
         * across a pixel.
         */
        int    u, v;
        int    newW, newH;

        u = (int)((width - 2) * s);
        v = (int)((height - 2) * t);

        newW = width - 2;
        newH = height - 2;

        if (s < 0 || u >= newW)
        {
            if (!repeating)
            {
                return false;
            }

            if (s < 0)
            {
                u = newW - 1 - ((-u) % newW);
            }
            else
            {
                u = u % newW;
            }
        } 

        if (t < 0 || v >= newH)
        {
            if (!repeating)
            {
                return false;
            }

            if (t < 0)
            {
                v = newH - 1 - ((-v) % newH);
            }
            else
            {
                v = v % newH;
            }
        }

        u += 1;
        v += 1;

        //
        // caclulate the differences
        //
        int     pixel;
        float   cU1, cU2, cV1, cV2;

        pixel = pix[u - 1 + v * width];
        cU1 = greyscale(getR(pixel), getG(pixel), getB(pixel));

        pixel = pix[u + 1 + v * width];
        cU2 = greyscale(getR(pixel), getG(pixel), getB(pixel));

        pixel = pix[u + (v - 1) * width];
        cV1 = greyscale(getR(pixel), getG(pixel), getB(pixel));

        pixel = pix[u + (v + 1) * width];
        cV2 = greyscale(getR(pixel), getG(pixel), getB(pixel));

        //
        // calculate perturbation
        //
        deltas[0] = (cU1 - cU2);
        deltas[1] = (cV1 - cV2);

        return true;
    }
}
