package org.autochthonous.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * A fog atmosphere. By default it affects all objects in the scene,
 * unless the choice is constrained by the using the addAffected method.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Fog color="1.0, 1.0, 1.0" fogFactor="0.001" rFactor="0.03"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *
 *     &lt;!ELEMENT PointLight EMPTY&gt;
 *     &lt;!ATTLIST PointLight
 *         color        %Color3f;  "1, 1, 1"
 *         fogFactor %Point3f;  "0.001"
 *         rFactor     %Float      "0.03"&gt;
 * </pre>
 * </code>
 */
public class Fog
    implements Atmosphere
{
    Color3f    col;
    float        fogFactor;
    float        rFactor;

    HashMap    affected = null;    // table of objects we affect

    /**
     * Create a fog object with the standard rFactor of 0.03.
     * 
     * @param col
     * @param fogFactor
     */
    public Fog(
        Color3f     col,
        float         fogFactor)
    {
        this(col, fogFactor, 0.03f);
    }

    public Fog(
        Color3f    col,
        float       fogFactor,
        float       rFactor)
    {
    	init(col, fogFactor, rFactor);
    }
    
    private Fog()
    {
    }

    public static Fog getInstance(
        RMLReader       in)
        throws IOException
    {
        ParametersMap   params = in.parameters();

        Fog  f = new Fog();

        f.init(
		    params.getColor("color", "1, 1, 1"),
            params.getFloat("fogFactor", "0.001"),
            params.getFloat("rFactor", "0.03"));

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return f;
    }

    private void init(
	    Color3f    col,
        float        fogFactor,
        float        rFactor)

    {
        this.col = new Color3f(col);
        this.fogFactor = fogFactor;
        this.rFactor = rFactor;
    }

    /**
     * transform the location of the fog.
     */
    public void transform(
        Transform   trans)
    {
        //
    }
    
    /**
     * Add an affected region to the fog. The fog will be limited in its affect
     * to the space occupied by the objects passed in to the this method.
     */
    public void addAffected(
        Traceable	obj)
    {
    }

    /**
     * Calculate the atmosphere data according to Beer's law of fog and haze.
     * 
     * @see org.autochthonous.graphics.rt.Atmosphere#getAtmosphereData(org.autochthonous.graphics.rt.Hit, org.autochthonous.graphics.rt.Illuminated)
     */
    public AtmosphereData getAtmosphereData(Hit hit, Illuminated world)
    {
		double fact = 1.0 - (1.0 - rFactor) * Math.exp((double)(-fogFactor * hit.t));
					
		return AtmosphereData.getInstance((float)fact, col);
    }
}
