package org.autochthonous.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * a heightfield.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Heightfield color="1, 1, 1" name="hfield.hf" vertex1="-1, -1, -1" vertex2="1, 1, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Heightfield EMPTY&gt;
 *     &lt;!ATTLIST Heightfield
 *         color   %Color3f;  "1, 1, 1"
 *         vertex1 %Point3f; "0, 0, 0"
 *         vertex2 %Point3f; "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class Heightfield
    extends Bounded
{
    private static final int    X = 0;
    private static final int    Y = 1;
    private static final int    Z = 2;

    public static final float   tolerance = (float)0.0001;

    private static final BBox   canonicalBBox = new BBox(0, 0, 0, 1, 1, 1);

    private static class BoxNormalCalculator 
            implements NormalCalculator
    {
        float   x, y, z;

        BoxNormalCalculator(
            float    x,
            float    y,
            float    z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }

        public void getN(
            Hit         hd,
            Vector3f    n)
        {
            n.set(x, y, z);
        }
    }

    /**
     * initialise a triangular polygon from the vertices a, b, and c.
     */
    private class Triangle
        implements Partitionable, ColorCalculator, NormalCalculator, TextureMapper
    {
        Vector3f    n = new Vector3f();
        BBox        bb = new BBox();
        float       cnst;
        float       m00, m01, m02, m10, m11, m12;
        byte        axis;
        Vector3f[]  norms = new Vector3f[3];

        void init(
            Vector3f     a,
            Vector3f     b,
            Vector3f     c)
        {
            Vector3f    v1 = new Vector3f(), v2 = new Vector3f();
            float       area, maxx, maxy, maxz, minx, miny, minz;
     
            v1.sub(b, a);
            v2.sub(c, b);
     
            n.cross(v1, v2);
     
            if (Math.abs(n.x) > Math.abs(n.y))
            {
                if (Math.abs(n.x) > Math.abs(n.z))
                {
                    axis = X;
                }
                else
                {
                    axis = Z;
                }
            }
            else
            {
                if (Math.abs(n.y) > Math.abs(n.z))
                {
                    axis = Y;
                }
                else
                {
                    axis = Z;
                }
            }
     
            if (n.x == 0.0 && n.y == 0.0 && n.z == 0.0)
            {
                throw new IllegalStateException("invalid polygon in heightfield\n");
            }
     
            //
            // as n isn't normalised the magnitude of the
            // normal in the part of 3 Space we ignore represents
            // the area of the projected triangle. Dividing through
            // by the component takes care of sign trouble we could
            // have in the intersection test so we don't worry about
            // taking the magnitude...
            //
            switch (axis)
            {
            case X:
                area = n.x;
                m00 = (a.y - c.y) / area;
                m10 = (b.y - a.y) / area;
     
                m01 = (a.z - c.z) / area;
                m11 = (b.z - a.z) / area;
     
                m02 = c.y * m01 - c.z * m00;
                m12 = a.y * m11 - a.z * m10;
                break;
            case Y:
                area = n.y;
                m00 = (c.x - a.x) / area;
                m10 = (a.x - b.x) / area;
     
                m01 = (c.z - a.z) / area;
                m11 = (a.z - b.z) / area;
     
                m02 = c.x * m01 - c.z * m00;
                m12 = a.x * m11 - a.z * m10;
                break;
            case Z:
                area = n.z;
                m00 = (a.x - c.x) / area;
                m10 = (b.x - a.x) / area;
     
                m01 = (a.y - c.y) / area;
                m11 = (b.y - a.y) / area;
     
                m02 = c.x * m01 - c.y * m00;
                m12 = a.x * m11 - a.y * m10;
                break;
            default:
                throw new IllegalStateException("bad axis in maketriangle.\n");
            }
     
            n.normalize();
     
            cnst = -n.dot(a);
     
            bb.set(a.x, a.y, a.z, a.x, a.y, a.z);
            bb.adjustFor(b.x, b.y, b.z);
            bb.adjustFor(c.x, c.y, c.z);
     
            bb.expand(tolerance);
        }

        public BBox bBox()
        {
            return bb;
        }

        void setNorms(
            Vector3f    n0,
            Vector3f    n1,
            Vector3f    n2)
        {
            this.norms[0] = n0;
            this.norms[1] = n1;
            this.norms[2] = n2;
        }

        /**
         * intersect a triangle, we calculate the barycentric coordinates
         * using the coefficients in the matrix we pre calculated in the
         * initialisation routine. This owes a bit to Didier Badouel from
         * his article in Graphics Gems.
         */
        boolean isTriangleIntersected(
            Ray        r)
        {
            float        t, pu, pv, a1, a2;
        
            t = n.dot(r.dir);
        
            if (t == 0.0)
            {
                return false;
            }
        
            t = -((n.x * r.org.x + n.y * r.org.y + n.z * r.org.z)
                            + cnst) / t;

            if (t < tolerance)
            {
                return false;
            }
        
            switch (axis)
            {
            case X:
                pu = r.org.y + t * r.dir.y;
                if (pu < bb.minY || pu > bb.maxY)
                {
                    return false;
                }
                pv = r.org.z + t * r.dir.z;
                if (pv < bb.minZ || pv > bb.maxZ)
                {
                    return false;
                }
                break;
            case Y:
                pu = r.org.x + t * r.dir.x;
                if (pu < bb.minX || pu > bb.maxX)
                {
                    return false;
                }
                pv = r.org.z + t * r.dir.z;
                if (pv < bb.minZ || pv > bb.maxZ)
                {
                    return false;
                }
                break;
            case Z:
                pu = r.org.x + t * r.dir.x;
                if (pu < bb.minX || pu > bb.maxX)
                {
                    return false;
                }
                pv = r.org.y + t * r.dir.y;
                if (pv < bb.minY || pv > bb.maxY)
                {
                    return false;
                }
                break;
            default:
                throw new IllegalStateException("bad axis in inttri.");
            }
        
            a1 = pv * m00 - pu * m01 + m02;
            if (a1 < 0.0 || a1 > 1.0)
            {
                return false;
            }
        
            a2 = pv * m10 - pu * m11 + m12;
            if (a2 < 0.0 || a1 + a2 > 1.0)
            {
                return false;
            }
        
            r.setHit(parent, t, this, this, this, shader); 

            return true;
        }

        public HitList intersections(
            Ray r)
        {
            if (isTriangleIntersected(r))
            {
                HitList list = HitList.getInstance();

                list.add(
                    Hit.getInstance(parent, r.hit.distance(),
                                this, this, this, r.hit.shader));

                return list;
            }

            return null;
        }

        public boolean isIntersected(
            Ray r)
        {
            return isTriangleIntersected(r);
        }

        public boolean isShadowing(
            Ray r)
        {
            return isTriangleIntersected(r);
        }

        public void getColor(
            Hit         hit,
            Color3f     c)
        {
            c.set(parent.getColor());
        }

        public void getN(
            Hit         hit,
            Vector3f    oN)
        {
            oN.set(this.n);
        }

        public boolean getST(
            Tuple3f l,
            float[] uv)
        {
            uv[S] = l.x;
            uv[T] = l.y;

            return true;
        }

        public Object clone()
            throws CloneNotSupportedException
        {
            return super.clone();
        }
    }

    private abstract class Node
    {
        protected BBox      nBB = new BBox();
        protected int       xStart, xEnd, yStart, yEnd;
        protected float     zMin, zMax;

        protected Node(
            int xStart,
            int xEnd,
            int yStart,
            int yEnd)
        {
            this.xStart = xStart;
            this.xEnd = xEnd;
            this.yStart = yStart;
            this.yEnd = yEnd;

            this.zMin = this.zMax = zVals[yStart][xStart];

            for (int y = yStart; y <= yEnd; y++)
            {
                float[] row = zVals[y];

                for (int x = xStart; x <= xEnd; x++)
                {
                    float   z = row[x];
                    if (z > this.zMax)
                    {
                        this.zMax = z;
                    }
                    else if (z < this.zMin)
                    {
                        this.zMin = z;
                    }
                }
            }

            nBB.set(((float)xStart / (float)(fieldsize - 1)),
                    ((float)yStart / (float)(fieldsize - 1)),
                    this.zMin,
                    ((float)xEnd / (float)(fieldsize - 1)),
                    ((float)yEnd / (float)(fieldsize - 1)),
                    this.zMax);
            nBB.expand(tolerance);
        }

        abstract HitList intersections(Ray r);

        abstract boolean isIntersected(Ray r);

        abstract boolean isShadowing(Ray r);
    }

    private class BranchNode
        extends Node
    {
        private int     type;
        private Node    left, right;

        BranchNode(
            int xStart,
            int xEnd,
            int yStart,
            int yEnd)
        {
            super(xStart, xEnd, yStart, yEnd);

            left = right = null;
        }

        void split()
        {
            if ((xEnd - xStart) > (yEnd - yStart))
            {
                int mid = (xEnd + xStart) / 2;

                if ((yEnd - yStart) == 1)
                {
                    if ((mid - xStart) == 1)
                    {
                        left = new LeafNode(xStart, mid, yStart, yEnd);
                    }
                    else
                    {
                        left = new BranchNode(xStart, mid, yStart, yEnd);
                    }

                    if ((xEnd - mid) == 1)
                    {
                        right = new LeafNode(mid, xEnd, yStart, yEnd);
                    }
                    else
                    {
                        right = new BranchNode(mid, xEnd, yStart, yEnd);
                    }
                }
                else
                {
                    left = new BranchNode(xStart, mid, yStart, yEnd);
                    right = new BranchNode(mid, xEnd, yStart, yEnd);
                }
            }
            else
            {
                int mid = (yEnd + yStart) / 2;

                if ((xEnd - xStart) == 1)
                {
                    if ((mid - yStart) == 1)
                    {
                        left = new LeafNode(xStart, xEnd, yStart, mid);
                    }
                    else
                    {
                        left = new BranchNode(xStart, xEnd, yStart, mid);
                    }

                    if ((yEnd - mid) == 1)
                    {
                        right = new LeafNode(xStart, xEnd, mid, yEnd);
                    }
                    else
                    {
                        right = new BranchNode(xStart, xEnd, mid, yEnd);
                    }
                }
                else
                {
                    left = new BranchNode(xStart, xEnd, yStart, mid);
                    right = new BranchNode(xStart, xEnd, mid, yEnd);
                }
            }
        }

        HitList intersections(
            Ray r)
        {
            if (nBB.isMissedBy(r))
            {
                return null;
            }

            if (left == null)
            {
                split();
            }
    
            //
            // check subtrees
            //
            HitList hits = null;

            if (type == X)
            {
                if (r.org.x <= (nBB.minX + nBB.maxX) / 2.0)
                {
                    hits = left.intersections(r);
                    if (r.dir.x >= 0.0)
                    {
                        if (hits != null)
                        {
                            hits.merge(right.intersections(r));
                        }
                        else
                        {
                            hits = right.intersections(r);
                        }
                    }
                }
                else
                {
                    hits = right.intersections(r);
                    if (r.dir.x <= 0.0)
                    {
                        if (hits != null)
                        {
                            hits.merge(left.intersections(r));
                        }
                        else
                        {
                            hits = left.intersections(r);
                        }
                    }
                }
            }
            else 
            {
                if (r.org.y <= (nBB.minY + nBB.maxY) / 2.0)
                {
                    hits = left.intersections(r);
                    if (r.dir.y >= 0.0)
                    {
                        if (hits != null)
                        {
                            hits.merge(right.intersections(r));
                        }
                        else
                        {
                            hits = right.intersections(r);
                        }
                    }
                }
                else
                {
                    hits = right.intersections(r);
                    if (r.dir.y <= 0.0)
                    {
                        if (hits != null)
                        {
                            hits.merge(left.intersections(r));
                        }
                        else
                        {
                            hits = left.intersections(r);
                        }
                    }
                }
            }

            return hits;
        }

        boolean isIntersected(
            Ray r)
        {
            if (nBB.isMissedBy(r))
            {
                return false;
            }

            if (left == null)
            {
                split();
            }
    
            //
            // check subtrees
            //
            boolean hit = false;

            if (type == X)
            {
                if (r.org.x <= (nBB.minX + nBB.maxX) / 2.0)
                {
                    hit = left.isIntersected(r);
                    if (!hit || r.dir.x >= 0.0)
                    {
                        return (hit | right.isIntersected(r));
                    }
                }
                else
                {
                    hit = right.isIntersected(r);
                    if (!hit || r.dir.x <= 0.0)
                    {
                        return (hit | left.isIntersected(r));
                    }
                }
            }
            else 
            {
                if (r.org.y <= (nBB.minY + nBB.maxY) / 2.0)
                {
                    hit = left.isIntersected(r);
                    if (!hit || r.dir.y >= 0.0)
                    {
                        return (hit | right.isIntersected(r));
                    }
                }
                else
                {
                    hit = right.isIntersected(r);
                    if (!hit || r.dir.y <= 0.0)
                    {
                        return (hit | left.isIntersected(r));
                    }
                }
            }

            return hit;
        }

        boolean isShadowing(
            Ray r)
        {
            if (nBB.isMissedBy(r))
            {
                return false;
            }

            if (left == null)
            {
                split();
            }
    
            //
            // check subtrees
            //
            boolean hit = false;

            if (type == X)
            {
                if (r.org.x <= (nBB.minX + nBB.maxX) / 2.0)
                {
                    hit = left.isShadowing(r);
                    if (!hit || r.dir.x >= 0.0)
                    {
                        return (hit | right.isShadowing(r));
                    }
                }
                else
                {
                    hit = right.isShadowing(r);
                    if (!hit || r.dir.x <= 0.0)
                    {
                        return (hit | left.isShadowing(r));
                    }
                }
            }
            else 
            {
                if (r.org.y <= (nBB.minY + nBB.maxY) / 2.0)
                {
                    hit = left.isShadowing(r);
                    if (!hit || r.dir.y >= 0.0)
                    {
                        return (hit | right.isShadowing(r));
                    }
                }
                else
                {
                    hit = right.isShadowing(r);
                    if (!hit || r.dir.y <= 0.0)
                    {
                        return (hit | left.isIntersected(r));
                    }
                }
            }

            return hit;
        }
    }

    private class LeafNode
        extends Node
    {
        Triangle    tri0 = new Triangle();
        Triangle    tri1 = new Triangle();
        Triangle    tri2 = new Triangle();
        Triangle    tri3 = new Triangle();

        Vector3f    cNorm = new Vector3f();

        LeafNode(
            int xStart,
            int xEnd,
            int yStart,
            int yEnd)
        {
            super(xStart, xEnd, yStart, yEnd);

            Vector3f    a = new Vector3f();
            Vector3f    b = new Vector3f();
            Vector3f    c = new Vector3f();
            Vector3f    d = new Vector3f();

            a.set(nBB.minX, nBB.minY, zVals[yStart][xStart]);
            b.set(nBB.maxX, nBB.minY, zVals[yStart][xStart + 1]);
            c.set(nBB.maxX, nBB.maxY, zVals[yStart + 1][xStart + 1]);
            d.set(nBB.minX, nBB.maxY, zVals[yStart + 1][xStart]);

            Vector3f    e = new Vector3f();
        
            e.x = (a.x + b.x + c.x + d.x) / 4.0f;
            e.y = (a.y + b.y + c.y + d.y) / 4.0f;
            e.z = (a.z + b.z + c.z + d.z) / 4.0f;
        
            tri0.init(a, b, e);
            tri1.init(b, c, e);
            tri2.init(c, d, e);
            tri3.init(d, a, e);

/*
            if (norms != null)
            {
                cNorm.x = norms[yStart][xStart].x + norms[ind + 1].x + norms[ind + xFieldsize].x + norms[ind + xFieldsize + 1].x;
                cNorm.y = norms[yStart][xStart].y + norms[ind + 1].y + norms[ind + xFieldsize].y + norms[ind + xFieldsize + 1].y;
                cNorm.z = norms[yStart][xStart].z + norms[ind + 1].z + norms[ind + xFieldsize].z + norms[ind + xFieldsize + 1].z;
                cNorm.normalize();

                tri0.setNorms(norms[ind], norms[ind + 1], cNorm);
                tri1.setNorms(norms[ind + 1], norms[ind + xFieldsize + 1], cNorm);
                tri2.setNorms(norms[ind + xFieldsize + 1], norms[ind + xFieldsize], cNorm);
                tri3.setNorms(norms[ind + xFieldsize], norms[ind], cNorm);
            }
*/
        }
    
        /**
         * check the objects in a leaf - returning a legal hit if there is
         * one, note in this case we (that of height fields), we don't have to
         * worry about over lapping objects, we just grab the closest one we hit
         * in the first leaf node we score a hit in.
         */
        HitList intersections(
            Ray   r)
        {
            HitList hits = tri0.intersections(r);

            if (hits != null)
            {
                hits.merge(tri1.intersections(r));
            }
            else
            {
                hits = tri1.intersections(r);
            }

            if (hits != null)
            {
                hits.merge(tri2.intersections(r));
            }
            else
            {
                hits = tri2.intersections(r);
            }

            if (hits != null)
            {
                hits.merge(tri3.intersections(r));
            }
            else
            {
                hits = tri3.intersections(r);
            }

            return hits;
        }

        boolean isIntersected(
            Ray r)
        {
            boolean hit = false;

            if (tri0.isIntersected(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri1.isIntersected(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri2.isIntersected(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri3.isIntersected(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            return hit;
        }

        boolean isShadowing(
            Ray r)
        {
            boolean hit = false;

            if (tri0.isShadowing(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri1.isShadowing(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri2.isShadowing(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            if (tri3.isShadowing(r))
            {
                hit = true;

                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }

            return hit;
        }
    }

    BBox        bb;

    //
    // z values, size, and various other values for the heightfield we
    // are intersecting
    //
    private int                 xFieldsize;
    private int                 yFieldsize;
    private int                 fieldsize;
    private int                 reset;
    private int                 freecount;
    private Triangle[]          tris;
    private Node[]              treenodes, freenodes;
    private Vector3f[]          norms;

    protected float[][]         zVals;
    protected int[]             pixVals;
    protected int               pixWidth, pixHeight;

    private Node                root;
    private Renderable          parent = this;

    protected Heightfield()
    {
    }

    protected Heightfield(
        Point3f     p1,
        Point3f     p2,
        float[][]   zVals)
    {
        initDim(p1, p2);
        initZ(zVals);
    }

    public static Heightfield getInstance(
        RMLReader       in)
        throws IOException
    {
        Heightfield b = new Heightfield();

        b.readObject(in);

        return b;
    }

    private void initDim(
        Point3f         p1,
        Point3f         p2)
    {
        double        minX, minY, minZ, maxX, maxY, maxZ;

        if (p1.x > p2.x)
        {
            maxX = p1.x;
            minX = p2.x;
        }
        else
        {
            maxX = p2.x;
            minX = p1.x;
        }

        if (p1.y > p2.y)
        {
            maxY = p1.y;
            minY = p2.y;
        }
        else
        {
            maxY = p2.y;
            minY = p1.y;
        }

        if (p1.z > p2.z)
        {
            maxZ = p1.z;
            minZ = p2.z;
        }
        else
        {
            maxZ = p2.z;
            minZ = p1.z;
        }

        this.setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale(maxX - minX, maxY - minY, maxZ - minZ);

        trans.translate(minX, minY, minZ);

        this.transform(trans);
    }

    protected void initZ(
        float[][]   zVals)
    {
        this.zVals = zVals;
        this.xFieldsize = zVals[0].length;
        this.yFieldsize = zVals.length;

        if (xFieldsize > yFieldsize)
        {
            fieldsize = xFieldsize;
        }
        else
        {
            fieldsize = yFieldsize;
        }

        if (xFieldsize > 2 && yFieldsize > 2)
        {
            root = new BranchNode(0, xFieldsize - 1, 0, yFieldsize - 1);
        }
        else
        {
            root = new LeafNode(0, xFieldsize - 1, 0, yFieldsize - 1);
        }
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        Point3f p1 = params.getPoint("vertex1", "0, 0, 0");
        Point3f p2 = params.getPoint("vertex2", "1, 1, 1");

        initDim(p1, p2);
    }

    public HitList intersections(
        Ray         r)
    {
        Ray         nr = this.createObjectSpaceRay(r);
        HitList     list = root.intersections(nr);

        nr.dispose();

        return list;
    }

    public boolean isIntersected(
        Ray    r)
    {
        Ray         nr = this.createObjectSpaceRay(r);
        boolean     hit = root.isIntersected(nr);

        if (hit)
        {
            r.setHit(this, nr.theHit.t,
                nr.theHit.colorCalc, nr.theHit.normCalc, nr.theHit.stMapper, nr.theHit.shader);
        }

        nr.dispose();

        return hit;
    }

    public boolean isShadowing(
        Ray    r)
    {
        Ray         nr = this.createObjectSpaceRay(r);
        boolean     hit = root.isIntersected(nr);

        if (hit)
        {
            r.setHit(this, nr.theHit.t,
                nr.theHit.colorCalc, nr.theHit.normCalc, nr.theHit.stMapper, nr.theHit.shader);
        }

        nr.dispose();

        return hit;
    }
}
