package org.autochthonous.graphics.rt.viewers;

import java.io.*;
import java.util.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * Class for the changing camera.
 * <p>
 * Sample RML for a changing camera is:
 * <code>
 * <pre>
 *      &lt;ChangingCamera&gt;
 *          &lt;Camera&gt;
 *              &lt;Perspective fov="45.0"/&gt;
 *              &lt;LookAt eye="0, 0, 5" reference="0, 0, 0" up="0, 1, 0"/&gt;
 *          &lt;/Camera&gt;
 *          &lt;Camera&gt;
 *              &lt;Perspective fov="45.0"/&gt;
 *              &lt;LookAt eye="0, 0, 6" reference="0, 0, 0" up="0, 1, 0"/&gt;
 *          &lt;/Camera&gt;
 *      &lt;/ChangingCamera&gt;
 * </pre>
 * </code>
 * <p>
 */
public abstract class ChangingCamera
    extends Viewer
{
    protected Camera[]    choices;
    protected float       viewFactor;

    public ChangingCamera(
        Camera[]  choices)
    {
        this.choices = choices;

        viewFactor = choices[0].viewFactor();

        for (int i = 0; i != choices.length; i++)
        {
            if (choices[i].viewFactor() > viewFactor)
            {
                viewFactor = choices[i].viewFactor();
            }
        }
    }

    public static Viewer getInstance(
        RMLReader   in)
        throws IOException
    {
        Object      obj;
        ArrayList   list = new ArrayList();

        while ((obj = in.readObject()) != null)
        {
            if (obj instanceof Camera)
            {
                list.add(obj);
            }
            else
            {
                throw new IOException("unknown object in ChangingCamera");
            }
        }

        return new LinearChangingCamera((Camera[])list.toArray(new Camera[list.size()]));
    }

    /**
     * configure a lense mask for this camera.
     */
    public void configureLenseMask(
        int        samples,
        boolean    jitter)
    {
        for (int i = 0; i != choices.length; i++)
        {
            choices[i].configureLenseMask(samples, jitter);
        }
    }

    /**
     * @param sampleNo number of this sample.
     * @param time time in the frame of this sample (0.0 to 1.0)
     * @param x x position on screen
     * @param y y position on screen
     */
    public abstract Ray generateRay(int sampleNo, float time, float x, float y);
}
